/*
**      Small-C Preprocessor Ver 1.1    26-Dec-87   cpp5.c
**      Nick Andrew & Jimi Hendrix
**
**      inline()      Read text from this or previous file
**                    Output in 'iline'
**      whiteline()   Remove comments and white space
**                    Input from mline, output to pline1
**      getw()        Get character into ch and next into nch
**      keepw(c)      Robustly put a character in *oline++;
**      whitesp(c)    Check for blank, tab or ff
*/

#include        "cpp.h"

/*
**      Inline ... Level 1-0 ... Read a line of text
*/

inline()
{
    for (;;) {
        if (!eof) {
            if (fgets(iline, LINEMAX, *input) != NULL) {
                char *eos;
                for (eos=iline;*eos;++eos) ;
                if (eos[-1]!='\n')
                    error("Input line too long");
                else
                    eos[-1] = 0;
                contin = (eos[-2]=='\\');
                break;
            }
            eof = 1;
        }
        /* if #asm in effect, complain */
        /* if comment in effect, complain */
        if (incomment) {
            error("EOF within comment");
            incomment = 0;
        }
        /* if string in effect, complain */
        fclose(*input);
        if (--input < fileptr) return;  /* no files outside */
        eof = 0;
    }

    fputs("Inline read: '",stderr);
    fputs(iline,stderr);
    fputs("'\n",stderr);
}

/*
**      Whiteline ... Level 2 ... Remove comments & whitespace
*/

whiteline()
{
    char    c;

    /* read line into mline */
    iline = mline;
    oline = pline1;
    ocount = 0;
    ch = 0;      /* this is wrong! or is it? */

    instring = enstring;
    incomment = encomment;

    inline();

    if (eof)
        return;

    ppline = (!instring && !incomment && *iline=='#');
    if (ppline) fputs("PPline!\n",stderr);

    enstring = instring;
    encomment = incomment;

    cptr = iline;

    while (getw()) {
        /* loop invariant: enstring & encomment are set if the current
           character is within a string or a comment
        */

        if (enstring) {

            while (ch!='"') {
                if (ch=='\\') {
                    keepw(ch);
                    getw();
                    if (ch) keepw(ch);
                    else break;
                }
                else if (ch==0) {
                    error("Missing close double quote");
                    enstring = 0;
                    break;
                }
                else {
                    keepw(ch);
                }
                getw();
            }

            if (ch=='"') {
                keepw(ch);
                enstring = 0;
            }
            continue;
        }

        if (encomment) {

            while ( (ch!='*' || nch!='/') && ch) {
                getw();
            }

            if (ch!=0) {
                getw();       /* get the "/" */
                keepw(' ');
                encomment = 0;
            }
            continue;
        }

        if (ch=='\'') {

            keepw(ch);
            getw();

            while (ch!='\'') {
                if (ch=='\\') {
                    keepw(ch);
                    getw();
                    if (ch) keepw(ch);
                    else continue;
                }
                else if (ch==0) {
                    error("Missing single quote");
                    keepw('\'');
                    break;
                }
                else {
                    keepw(ch);
                }
                getw();
            }

            if (ch=='\'') {
                keepw(ch);
            }
            continue;
        }

        if (whitesp(ch)) {
            if (!whitesp(nch)) keepw(' ');
            continue;
        }

        if (ch=='"') {
            enstring = 1;
        }
        else if (ch=='/' && nch=='*') {
            encomment = 1;
            getw();          /* get the '*' */
            continue;
        }

        keepw(ch);
    }

    if (ocount >= LINEMAX)
        error("line too long (whiteline)");

    keepw(0);

    fputs("Whiteline kept: '",stderr);
    fputs(pline1,stderr);
    fputs("'\n",stderr);
}

keepw(c)
char    c;
{
        if (ocount < LINEMAX) {
            *(oline++) = c;
            ++ocount;
        }
}

whitesp(c)
int     c;
{
    return (c==' ') | (c=='\t');
}

int     getw()
{
    ch = *cptr;
    if (ch==0)
        nch=0;
    else
        nch = *++cptr;
     return ch;
}

