/*
**      Small-C Preprocessor Ver 1.1    26-Dec-87   cpp6.c
**      Nick Andrew
**      C expression evaluator for C preprocessor
** uses an operator precedence table ref dragon book, Sec 4.6
*/

#define STACK    200    /* length of stack in integers */
#define MATRIX    24    /* 24 operators, (, ), ID, $ */

#define NUMBER    -1
#define PLUS       0
#define MINUS      1
#define TIMES      2
#define DIVIDE     3
#define MOD        4
#define AND        5
#define OR         6
#define XOR        7
#define SHL        8
#define SHR        9
#define EQ        10
#define NOTEQ     11
#define LESS      12
#define GRTR      13
#define LESSEQ    14
#define GRTREQ    15
#define UNMINUS   16
#define UNNOT     17
#define ID        18
#define PAREN_O   19
#define PAREN_C   20
#define EOS       21
#define TERN_Q    22
#define TERN_C    23

#define N          0    /* ' ' error condition */
#define R          1    /* '>' takes precedence over */
#define L          2    /* '<' yields precedence to */
#define E          3    /* '=' has equal precedence */

int *rel;

int rel1[MATRIX*MATRIX/2] = {

/*        +  -  *  /  %  &  |  ^  << 
>> == != <  >  <= >= U- U~ ID (  )  $  ?  :  */

/* + */   R, R, L, L, L, R, R, R, R,  
R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* - */   R, R, L, L, L, R, R, R, R,  
R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* * */   R, R, R, R, R, R, R, R, R,  
R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* / */   R, R, R, R, R, R, R, R, R,  
R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* % */   R, R, R, R, R, R, R, R, R,  
R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* & */   L, L, L, L, L, R, R, R, L,  
L, L, L, L, L, L, L, L, L, L, L, R, R, R, R, 
/* | */   L, L, L, L, L, L, R, L, L,  
L, L, L, L, L, L, L, L, L, L, L, R, R, R, R, 
/* ^ */   L, L, L, L, L, L, R, R, L,  
L, L, L, L, L, L, L, L, L, L, L, R, R, R, R, 
/* <<*/   L, L, L, L, L, R, R, R, R,  
R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* >>*/   L, L, L, L, L, R, R, R, R,  
R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* ==*/   L, L, L, L, L, R, R, R, L,  
L, R, R, L, L, L, L, L, L, L, L, R, R, R, R, 
/* !=*/   L, L, L, L, L, R, R, R, L,  
L, R, R, L, L, L, L, L, L, L, L, R, R, R, R 

};

int  rel2[MATRIX*MATRIX/2] = {

/* < */   L, L, L, L, L, R, R, R, L,  
          L, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* > */   L, L, L, L, L, R, R, R, L,  
          L, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* <=*/   L, L, L, L, L, R, R, R, L,  
          L, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* >=*/   L, L, L, L, L, R, R, R, L,  
          L, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* U-*/   R, R, R, R, R, R, R, R, R,  
          R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* U~*/   R, R, R, R, R, R, R, R, R,  
          R, R, R, R, R, R, R, L, L, L, L, R, R, R, R, 
/* ID*/   R, R, R, R, R, R, R, R, R,  
          R, R, R, R, R, R, R, R, R, N, N, R, R, R, R, 
/* ( */   L, L, L, L, L, L, L, L, L,  
          L, L, L, L, L, L, L, L, L, L, L, E, N, L, N, 
/* ) */   R, R, R, R, R, R, R, R, R,  
          R, R, R, R, R, R, R, R, R, N, N, R, R, R, R, 
/* $ */   L, L, L, L, L, L, L, L, L,  
          L, L, L, L, L, L, L, L, L, L, L, N, N, L, N, 
/* ? */   L, L, L, L, L, L, L, L, L,  
          L, L, L, L, L, L, L, L, L, L, L, N, N, L, E, 
/* : */   L, L, L, L, L, L, L, L, L,  
          L, L, L, L, L, L, L, L, L, L, L, R, R, L, R 

};

int  stack[STACK],*sp,*topterm;
int  product1[5],product2[5];       /* most complex is E?E:E */
int  token,tokval,result,eerror;
char estr[LINESIZE];

/*
**  Evaluate a #if control line to 0 or non-zero
*/

int     eval_if()
{
    eerror = 0;
    rel = rel1;
    topterm = sp = stack;

    /* do macro substitution on the #if line */
    macsub();

    iline = pline2;
    bump(0);
    match("#if");
    blanks();
    cptr = lptr;

    push(EOS);
    exprval();
    if (sp != stack + 2) {
        eerror = 1;
    }

    result=pop();
    if (eerror) {
        error("Invalid #if expression (1)");
        return 0;
    }

    outline("Result is ",result);
    return result;
}

outline(s,n)
char *s;
int  n;
{
    char buff[8];
    fputs(s,stdout);
    itoa(n,buff);
    fputs(buff,stdout);
    fputs("\n",stdout);
}

push(n)
int  n;
{
    outline("Pushing to stack: ",n);
    if (sp >= stack + (STACK - 1)) {
        eerror = 1;
        error("#if Expression too complex");
        return;
    }

    if (n == NUMBER || n == ID) *(sp++) = tokval;

    if (n != NUMBER) topterm = sp;

    *(sp++) = n;
}

int  pop() {
    int type;
    outline("Popping from stack: ",*sp);
    if (sp <= stack) {
        eerror = 1;
        return 0;
    }

    type = *(--sp);
    if (type == NUMBER || type == ID)
       tokval = *(--sp);

    topterm = sp;
    while (*topterm == NUMBER) topterm -= 2;

    return type;
}

/*
**   Exprval ... Parse and evaluate constant expression
*/

exprval()
{
    int     token,relation,nterms,popped;

    token = gettok();
    for (;;) {
        if (eerror) return;
        if (*topterm == EOS && token == EOS) return;

        relation = rel(*topterm,token);

        if (relation == L || relation == E) {
            push(token);
            token = gettok();
        }
        else if (relation == R) {
            nterms = 0;
            do {
                product1[nterms] = popped = pop();
                product2[nterms] = tokval;
            } while (nterms++ < 5 && rel(*topterm,popped)!=L);
            reduce(nterms);
        }
        else {
            eerror = 1;
        }
    }
}

/*
**   Reduce ... reduce a production of N terms. The terms are
**   stored backwards in product1 & values in product2
*/

reduce(nterms)
int    nterms;
{
    int   val1,val2;

    if (nterms > 5) {
        eerror = 1;
        error("More than 5 terms in a production");
        return;
    }

    switch(nterms) {

    case 1 :                        /* E -> id */
        token = product1[0];
        if (token == ID) {
            tokval = product2[0];
            push(NUMBER);
        } else {
            eerror = 1;
        }
        return;

    case 2 :                        /* E -> ~ E */
        token = product1[1];
        tokval = product2[0];
        switch(token) {
        case UNMINUS :
            tokval = -tokval;
            break;
        case UNNOT :
            tokval = ~tokval;
            break;
        default :
            eerror = 1;
            return;
        }
        push(NUMBER);
        return;

    case 3 :                  /* E -> EAE or E -> (E) */
        if (product1[1] == NUMBER) {
            if (product1[0] == PAREN_C && product1[2] == PAREN_O) {
                tokval = product2[1];
                push(NUMBER);
            } else {
                eerror = 1;
            }
            return;
        }

        token = product1[1];
        val1 = product2[2];
        val2 = product2[0];

        if (product1[2] != NUMBER || product1[0] != NUMBER) {
            eerror = 1;
            return;
        }

        switch(token) {
        case PLUS :
            tokval = val1 + val2;
            break;
        case MINUS :
            tokval = val1 - val2;
            break;
        case TIMES :
            tokval = val1 * val2;
            break;
        case DIVIDE :
            tokval = val1 / val2;
            break;
        case MOD :
            tokval = val1 % val2;
            break;
        case AND :
            tokval = val1 & val2;
            break;
        case OR :
            tokval = val1 | val2;
            break;
        case XOR :
            tokval = val1 ^ val2;
            break;
        case SHL :
            tokval = val1 << val2;
            break;
        case SHR :
            tokval = val1 >> val2;
            break;
        case EQUAL :
            tokval = val1 == val2;
            break;
        case NOTEQ :
            tokval = val1 != val2;
            break;
        case LESS :
            tokval = val1 < val2;
            break;
        case GRTR :
            tokval = val1 > val2;
            break;
        case LESSEQ :
            tokval = val1 <= val2;
            break;
        case GRTREQ :
            tokval = val1 >= val2;
            break;
        default :
            eerror = 1;
            return;
        }
        push(NUMBER);

    case 4 :
        eerror = 1;
        error("No 4-term productions");
        return;

    case 5 :
        if (product1[1] != TERN_C || product1[4] != TERN_Q) {
            eerror = 1;
            error("Not the ternary operator ?:");
            return;
        }

        if (product1[0] != NUMBER
            || product1[2] != NUMBER
            || product1[4] != NUMBER) {
            eerror = 1;
            return;
        }

        /* we can't compile ?: (hee hee) */
        if (product2[4])
            tokval = product2[2];
        else
            tokval = product2[0];

        push(NUMBER);
        return;

    default : eerror = 1;
    }
}

/*
**   Gettok ... Get the next input token.
**   Recognise unary minus.
**   Determine the value of character and integer constants
**   This will be/is/was a real mother to write
**   Return EOS on end of line
*/

int     gettok()
{
}
